import requests
import json
from exceptions.exceptions import InvalidAlertType
from utils.time import convert_utc_time_to_local_time


class Alert(object):
    ALERT_DESCRIPTION = None

    def __init__(self, alert_id) -> None:
        self.alert_id = alert_id

    @staticmethod
    def create_alert_from_row(alert_row: list) -> 'Alert':
        alert_id, detected_at, database_name, schema_name, table_name, column_name, alert_type, sub_type, \
            alert_description = alert_row
        if alert_type == 'schema_change':
            return SchemaChangeAlert(alert_id, database_name, schema_name, table_name, detected_at, sub_type,
                                     alert_description)
        else:
            raise InvalidAlertType(f'Got invalid alert type - {alert_type}')

    @staticmethod
    def send(webhook: str, data: dict, content_types: str = "application/json"):
        requests.post(url=webhook, headers={'Content-type': content_types}, data=json.dumps(data))

    def to_slack_message(self) -> dict:
        pass

    def send_to_slack(self, webhook: str):
        data = self.to_slack_message()
        self.send(webhook, data)

    @property
    def id(self) -> str:
        return self.alert_id


class SchemaChangeAlert(Alert):
    ALERT_DESCRIPTION = "Schema change detected"

    def __init__(self, alert_id, database_name, schema_name, table_name, detected_at, sub_type, description) -> None:
        super().__init__(alert_id)
        self.table_name = '.'.join([database_name, schema_name, table_name]).lower()
        self.detected_at = convert_utc_time_to_local_time(detected_at).strftime('%Y-%m-%d %H:%M:%S')
        self.description = description[0].upper() + description[1:].lower()
        self.change_type = ' '.join([word[0].upper() + word[1:] for word in sub_type.split('_')])

    def to_slack_message(self) -> dict:
        return {
            "blocks": [
                {
                    "type": "divider"
                },
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f":small_red_triangle:*{self.ALERT_DESCRIPTION}*"
                    }
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": f">*Table:*\n>{self.table_name}"
                        },
                        {
                            "type": "mrkdwn",
                            "text": f"*When:*\n{self.detected_at}"
                        }
                    ]
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": f">*Change Type:*\n>{self.change_type}"
                        },
                        {
                            "type": "mrkdwn",
                            "text": f"*Description:*\n{self.description}"
                        }
                    ]
                }
            ]
        }







from datetime import datetime, timedelta
from typing import Optional, Union
from lineage.bigquery_query import BigQueryQuery
from exceptions.exceptions import SerializationError
from lineage.query import Query
from lineage.query_history_stats import QueryHistoryStats
from lineage.snowflake_query import SnowflakeQuery
from utils.env_vars import is_flight_mode_on
import json
import os


class QueryHistory(object):

    INFORMATION_SCHEMA_QUERY_HISTORY = None
    PLATFORM_TYPE = None
    SUCCESS_QUERIES_FILE = './latest_query_history.json'
    FAILED_QUERIES_FILE = './failed_queries.json'

    def __init__(self, con, dbs: str, should_export_query_history: bool = True, full_table_names: bool = True) -> None:
        self._con = con
        self._dbs = self.str_to_list(dbs)
        self._should_export_query_history = should_export_query_history
        self._full_table_names = full_table_names
        self._query_history_stats = QueryHistoryStats()
        self.success_queries = []
        self.failed_queries = []

    @staticmethod
    def str_to_list(dbs_str: str) -> Union[list, None]:
        if dbs_str is None:
            return None

        return [db.strip() for db in dbs_str.split(',')]

    @staticmethod
    def _normalize_database_name(db: str) -> str:
        return db.lower().replace('-', '_').replace(' ', '').replace('`', '').replace('"', '')

    @staticmethod
    def serialize_queries_to_file(filename: str, queries: [Query]) -> None:
        with open(filename, 'w') as queries_file:
            serialized_queries = []
            for query in queries:
                serialized_queries.append(query.to_dict())
            json.dump(serialized_queries, queries_file)

    def _serialize_query_history(self) -> None:
        if self._should_export_query_history:
            self.serialize_queries_to_file(self.SUCCESS_QUERIES_FILE, self.success_queries)
            self.serialize_queries_to_file(self.FAILED_QUERIES_FILE, self.failed_queries)

    def _deserialize_query_history(self) -> [Query]:
        if os.path.exists(self.SUCCESS_QUERIES_FILE):
            with open(self.SUCCESS_QUERIES_FILE, 'r') as query_history_file:
                queries = json.load(query_history_file)
                for query_dict in queries:
                    platform_type = query_dict.pop('platform_type')
                    if platform_type == SnowflakeQuery.PLATFORM_TYPE:
                        query = SnowflakeQuery.from_dict(query_dict)
                    elif platform_type == BigQueryQuery.PLATFORM_TYPE:
                        query = BigQueryQuery.from_dict(query_dict)
                    else:
                        raise SerializationError(f'Invalid platform type - {platform_type}')
                    self.add_query(query)

    @staticmethod
    def _include_end_date(end_date: datetime) -> Optional[datetime]:
        if end_date is not None and (end_date.hour, end_date.minute, end_date.second) == (0, 0, 0):
            return end_date + timedelta(hours=23, minutes=59, seconds=59)

        return end_date

    def add_query(self, query: Query):
        if query.parse(self._full_table_names):
            self.success_queries.append(query)
        else:
            self.failed_queries.append(query)
        self._query_history_stats.update_stats(query.query_context)

    def extract_queries(self, start_date: datetime, end_date: datetime) -> [Query]:
        if is_flight_mode_on():
            self._deserialize_query_history()
        else:
            self._query_history_table(start_date, end_date)
            self._serialize_query_history()

        return self.success_queries

    def _query_history_table(self, start_date: datetime, end_date: datetime) -> [Query]:
        pass

    def properties(self) -> dict:
        failed_queries_count = len(self.failed_queries)
        success_queries_count = len(self.success_queries)
        queries_count = success_queries_count + failed_queries_count
        query_history_properties = {'query_history_properties': {'failed_queries': failed_queries_count,
                                                                 'success_queries': success_queries_count,
                                                                 'queries_count': queries_count,
                                                                 'platform_type': self.PLATFORM_TYPE}}
        query_history_properties.update({'query_stats': self._query_history_stats.to_dict()})
        return query_history_properties




import click
import os
from os.path import expanduser

from utils.package import get_package_version
from tracking.anonymous_tracking import AnonymousTracking, track_cli_start, track_cli_exception, track_cli_end
from utils.dbt import get_profile_name_from_dbt_project, is_dbt_installed
from utils.ordered_yaml import OrderedYaml
from config.config import Config
from monitor.data_monitoring import DataMonitoring

yaml = OrderedYaml()


def get_cli_properties() -> dict:

    click_context = click.get_current_context()
    if click_context is None:
        return dict()

    params = click_context.params
    if params is None:
        return dict()

    reload_monitoring_configuration = params.get('reload_monitoring_configuration')
    update_dbt_package = params.get('update_dbt_package')
    full_refresh_dbt_package = params.get('full_refresh_dbt_package')

    return {'reload_monitoring_configuration': reload_monitoring_configuration,
            'update_dbt_package': update_dbt_package,
            'full_refresh_dbt_package': full_refresh_dbt_package,
            'dbt_installed': is_dbt_installed(),
            'version': get_package_version()}



@click.command()
@click.option(
    '--reload-monitoring-configuration', '-r',
    type=bool,
    default=False,
    help="edr automatically uploads the monitoring configuration if it does not exist in your data warehouse, "
         "use this flag if you changed your configuration and want to reload it."
)
@click.option(
    '--config-dir', '-c',
    type=str,
    default=os.path.join(expanduser('~'), '.edr'),
    help="Global settings for edr are configured in a config.yml file in this directory "
         "(if your config dir is HOME_DIR/.edr, no need to provide this parameter as we use it as default)."
)
@click.option(
    '--profiles-dir', '-p',
    type=str,
    default=os.path.join(expanduser('~'), '.dbt'),
    help="Specify your profiles dir where a profiles.yml is located, this could be a dbt profiles dir "
         "(if your profiles dir is HOME_DIR/.dbt, no need to provide this parameter as we use it as default).",
)
@click.option(
    '--update-dbt-package', '-u',
    type=bool,
    default=False,
    help="Force downloading the latest version of the edr internal dbt package (usually this is not needed, "
         "see documentation to learn more)."
)
@click.option(
    '--full-refresh-dbt-package', '-f',
    type=bool,
    default=False,
    help="Force running a full refresh of all incremental models in the edr dbt package (usually this is not needed, "
         "see documentation to learn more)."
)
@click.pass_context
def monitor(ctx, reload_monitoring_configuration, config_dir, profiles_dir, update_dbt_package, full_refresh_dbt_package):
    click.echo(f"Any feedback and suggestions are welcomed! join our community here - "
               f"https://bit.ly/slack-elementary\n")
    config = Config(config_dir, profiles_dir, get_profile_name_from_dbt_project(DataMonitoring.DBT_PROJECT_PATH))
    anonymous_tracking = AnonymousTracking(config)
    track_cli_start(anonymous_tracking, 'monitor', get_cli_properties(), ctx.command.name)
    try:
        data_monitoring = DataMonitoring.create_data_monitoring(config)
        data_monitoring.run(reload_monitoring_configuration, update_dbt_package, full_refresh_dbt_package)
        track_cli_end(anonymous_tracking, 'monitor', data_monitoring.properties(), ctx.command.name)
    except Exception as exc:
        track_cli_exception(anonymous_tracking, 'monitor', exc, ctx.command.name)
        raise


if __name__ == "__main__":
    monitor()



import click
from pyfiglet import Figlet
import os
from os.path import expanduser

from config.config import Config
from tracking.anonymous_tracking import AnonymousTracking, track_cli_help

f = Figlet(font='slant')
print(f.renderText('Elementary'))

root_folder = os.path.join(os.path.dirname(__file__), '..')
modules = ['lineage', 'monitor']


class ElementaryCLI(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for module in modules:
            rv.append(module)
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(root_folder, name, 'cli.py')
        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            eval(code, ns, ns)
        return ns[name]

    def format_help(self, ctx, formatter):
        try:
            config = Config(config_dir=os.path.join(expanduser('~'), '.edr'),
                            profiles_dir=os.path.join(expanduser('~'), '.dbt'),
                            profile_name='elementary')
            anonymous_tracking = AnonymousTracking(config)
            track_cli_help(anonymous_tracking)
        except Exception:
            pass
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)


cli = ElementaryCLI(help='Open source data reliability solution (https://docs.elementary-data.com/)')

if __name__ == '__main__':
    cli()



import os
from typing import Any
import snowflake.connector.errors
from exceptions.exceptions import ConfigError
from monitor.alerts import Alert
from monitor.dbt_runner import DbtRunner
from config.config import Config
from utils.dbt import extract_credentials_and_data_from_profiles, get_profile_name_from_dbt_project, \
    get_snowflake_client
from utils.log import get_logger

logger = get_logger(__name__)
FILE_DIR = os.path.dirname(__file__)


class DataMonitoring(object):
    DBT_PACKAGE_NAME = 'elementary_data_reliability'
    DBT_PROJECT_PATH = os.path.join(FILE_DIR, 'dbt_project')
    # Compatibility for previous dbt versions
    DBT_PROJECT_MODULES_PATH = os.path.join(DBT_PROJECT_PATH, 'dbt_modules', DBT_PACKAGE_NAME)
    DBT_PROJECT_PACKAGES_PATH = os.path.join(DBT_PROJECT_PATH, 'dbt_packages', DBT_PACKAGE_NAME)
    DBT_PROJECT_SEEDS_PATH = os.path.join(DBT_PROJECT_PATH, 'data')

    MONITORING_CONFIGURATION = 'monitoring_configuration'

    SELECT_NEW_ALERTS_QUERY = """
        SELECT alert_id, detected_at, database_name, schema_name, table_name, column_name, alert_type, sub_type, 
               alert_description
            FROM ALERTS
            WHERE alert_sent = FALSE;
    """

    UPDATE_SENT_ALERTS = """
        UPDATE ALERTS set alert_sent = TRUE
            WHERE alert_id IN (%s); 
    """

    COUNT_ROWS_QUERY = None

    def __init__(self, config: Config, db_connection: Any) -> None:
        self.config = config
        self.dbt_runner = DbtRunner(self.DBT_PROJECT_PATH, self.config.profiles_dir)
        self.db_connection = db_connection
        self.execution_properties = {}

    @staticmethod
    def create_data_monitoring(config: Config) -> 'DataMonitoring':
        if config.platform == 'snowflake':
            snowflake_conn = get_snowflake_client(config.credentials, server_side_binding=False)
            return SnowflakeDataMonitoring(config, snowflake_conn)
        else:
            raise ConfigError("Unsupported platform")

    def _dbt_package_exists(self) -> bool:
        return os.path.exists(self.DBT_PROJECT_PACKAGES_PATH) or os.path.exists(self.DBT_PROJECT_MODULES_PATH)

    def _upload_configuration(self) -> bool:
        target_csv_dir = self.DBT_PROJECT_SEEDS_PATH
        if not os.path.exists(target_csv_dir):
            os.makedirs(target_csv_dir)

        monitoring_config_csv_path = os.path.join(target_csv_dir, f'{self.MONITORING_CONFIGURATION}.csv')
        csv_row_count = self.config.monitoring_configuration_in_dbt_sources_to_csv(monitoring_config_csv_path)
        self.execution_properties['configuration_csv_row_count'] = csv_row_count

        return self.dbt_runner.seed()

    def _run_query(self, query: str, params: tuple = None) -> list:
        pass

    def monitoring_configuration_exists(self) -> bool:
        pass

    def _update_sent_alerts(self, alert_ids) -> None:
        results = self._run_query(self.UPDATE_SENT_ALERTS, (alert_ids,))
        logger.debug(f"Updated sent alerts -\n{str(results)}")

    def _query_alerts(self) -> list:
        alert_rows = self._run_query(self.SELECT_NEW_ALERTS_QUERY)
        self.execution_properties['alert_rows'] = len(alert_rows)
        alerts = []
        for alert_row in alert_rows:
            alerts.append(Alert.create_alert_from_row(alert_row))
        return alerts

    def _send_to_slack(self, alerts: [Alert]) -> None:
        slack_webhook = self.config.slack_notification_webhook
        if slack_webhook is not None:
            sent_alerts = []
            for alert in alerts:
                alert.send_to_slack(slack_webhook)
                sent_alerts.append(alert.id)

            sent_alert_count = len(sent_alerts)
            self.execution_properties['sent_alert_count'] = sent_alert_count
            if sent_alert_count > 0:
                self._update_sent_alerts(sent_alerts)
        else:
            logger.info("Alerts found but slack webhook is not configured (see documentation on how to configure "
                        "a slack webhook)")

    def _download_dbt_package_if_needed(self, force_update_dbt_packages: bool):
        internal_dbt_package_exists = self._dbt_package_exists()
        self.execution_properties['dbt_package_exists'] = internal_dbt_package_exists
        self.execution_properties['force_update_dbt_packages'] = force_update_dbt_packages
        if not internal_dbt_package_exists or force_update_dbt_packages:
            logger.info("Downloading edr internal dbt package")
            package_downloaded = self.dbt_runner.deps()
            self.execution_properties['package_downloaded'] = package_downloaded
            if not package_downloaded:
                logger.info('Could not download internal dbt package')
                return

    def _upload_monitoring_configuration_if_needed(self, reload_monitoring_configuration: bool):
        monitoring_configuration_exists = self.monitoring_configuration_exists()
        self.execution_properties['monitoring_configuration_exists'] = monitoring_configuration_exists
        self.execution_properties['reload_monitoring_configuration'] = reload_monitoring_configuration
        if not monitoring_configuration_exists or reload_monitoring_configuration:
            logger.info("Uploading monitoring configuration")
            config_updated = self._upload_configuration()
            self.execution_properties['config_uploaded'] = config_updated
            if not config_updated:
                logger.info('Could not upload monitoring configuration')
                return

    def _send_alerts(self):
        alerts = self._query_alerts()
        alert_count = len(alerts)
        self.execution_properties['alert_count'] = alert_count
        if alert_count > 0:
            self._send_to_slack(alerts)

    def run(self, reload_monitoring_configuration: bool = False, force_update_dbt_package: bool = False,
            dbt_full_refresh: bool = False) -> None:

        self._download_dbt_package_if_needed(force_update_dbt_package)
        self._upload_monitoring_configuration_if_needed(reload_monitoring_configuration)

        logger.info("Running dbt snapshots to detect changes in schemas")
        success = self.dbt_runner.snapshot()
        self.execution_properties['snapshot_success'] = success
        if not success:
            logger.info('Could not run dbt snapshot successfully')
            return

        logger.info("Running edr internal dbt package")
        success = self.dbt_runner.run(model=self.DBT_PACKAGE_NAME, full_refresh=dbt_full_refresh)
        self.execution_properties['run_success'] = success
        if not success:
            logger.info('Could not run dbt run successfully')
            return

        self._send_alerts()

    def properties(self):
        data_monitoring_properties = {'data_monitoring_properties': self.execution_properties}
        return data_monitoring_properties


class SnowflakeDataMonitoring(DataMonitoring):

    COUNT_ROWS_QUERY = """
    SELECT count(*) FROM IDENTIFIER(%s)
    """

    def __init__(self, config: 'Config', db_connection: Any):
        super().__init__(config, db_connection)

    def _run_query(self, query: str, params: tuple = None) -> list:
        with self.db_connection.cursor() as cursor:
            if params is not None:
                cursor.execute(query, params)
            else:
                cursor.execute(query)

            results = cursor.fetchall()
            return results

    def _count_rows_in_table(self, table_name) -> int:
        try:
            results = self._run_query(self.COUNT_ROWS_QUERY,
                                      (table_name,))
            if len(results) == 1:
                return results[0][0]
        except snowflake.connector.errors.ProgrammingError:
            pass

        return 0

    def monitoring_configuration_exists(self) -> bool:
        if self._count_rows_in_table(self.MONITORING_CONFIGURATION) > 0:
            return True

        return False





import itertools
import os.path
from typing import Optional, Union
import networkx as nx
from collections import defaultdict
from exceptions.exceptions import ConfigError
from pyvis.network import Network
import webbrowser
from lineage.query import Query
from utils.log import get_logger
from alive_progress import alive_it
import json
import pkg_resources

logger = get_logger(__name__)

GRAPH_VISUALIZATION_OPTIONS = """{
            "edges": {
                "color": {
                  "color": "rgba(23,107,215,1)",
                  "highlight": "rgba(23,107,215,1)",
                  "hover": "rgba(23,107,215,1)",
                  "inherit": false
                },
                "dashes": true,
                "smooth": {
                    "type": "continuous",
                    "forceDirection": "none"
                }
            },
            "layout": {
                "hierarchical": {
                    "enabled": true,
                    "levelSeparation": 485,
                    "nodeSpacing": 100,
                    "treeSpacing": 100,
                    "blockShifting": false,
                    "edgeMinimization": true,
                    "parentCentralization": false,
                    "direction": "LR",
                    "sortMethod": "directed"
                }
            },
            "interaction": {
                "hover": true,
                "navigationButtons": true,
                "multiselect": true,
                "keyboard": {
                    "enabled": true
                }
            },
            "physics": {
                "enabled": false,
                "hierarchicalRepulsion": {
                    "centralGravity": 0
                },
                "minVelocity": 0.75,
                "solver": "hierarchicalRepulsion"
            }
}"""


class LineageGraph(object):
    UPSTREAM_DIRECTION = 'upstream'
    DOWNSTREAM_DIRECTION = 'downstream'
    BOTH_DIRECTIONS = 'both'
    SELECTED_NODE_COLOR = '#0925C7'
    SELECTED_NODE_TITLE = 'Selected table<br/><br/>'

    LINEAGE_GRAPH_FILE_MAME = 'lineage_graph.gpickle'
    LINEAGE_GRAPH_ATTRIBUTES_FILE_NAME = 'graph_attributes.json'

    def __init__(self, show_isolated_nodes: bool = False) -> None:
        self._lineage_graph = nx.DiGraph()
        self._show_isolated_nodes = show_isolated_nodes
        self._graph_attributes = defaultdict(lambda: None)

    def _update_lineage_graph(self, query: Query) -> None:
        # Handle drop tables, if they exist in the statement
        for dropped_table_name in query.dropped_tables:
            self._remove_node(dropped_table_name)

        # Handle rename tables
        for old_table_name, new_table_name in query.renamed_tables:
            self._rename_node(old_table_name, new_table_name)

        self._add_nodes_and_edges(query.source_tables, query.target_tables, query.get_context_as_html())

    def _add_nodes_and_edges(self, sources: {str}, targets: {str}, query_context_html: str) -> None:
        if None in sources:
            sources.remove(None)
        if None in targets:
            targets.remove(None)

        if not sources and not targets:
            return

        if len(sources) > 0 and len(targets) == 0:
            if self._show_isolated_nodes:
                self._lineage_graph.add_nodes_from(sources)
        elif len(targets) > 0 and len(sources) == 0:
            if self._show_isolated_nodes:
                for target_node in targets:
                    self._lineage_graph.add_node(target_node)
                    self._graph_attributes[target_node] = query_context_html
        else:
            self._lineage_graph.add_nodes_from(sources)
            for target_node in targets:
                self._lineage_graph.add_node(target_node)
                self._graph_attributes[target_node] = query_context_html
            for source, target in itertools.product(sources, targets):
                self._lineage_graph.add_edge(source, target)

    def _rename_node(self, old_node: str, new_node: str) -> None:
        if old_node is None or new_node is None:
            return

        if self._lineage_graph.has_node(old_node):
            # Rename in place instead of copying the entire lineage graph
            nx.relabel_nodes(self._lineage_graph, {old_node: new_node}, copy=False)
            if old_node in self._graph_attributes:
                old_node_attributes = self._graph_attributes[old_node]
                del self._graph_attributes[old_node]
                self._graph_attributes[new_node] = old_node_attributes

    def _remove_node(self, node: str) -> None:
        # First let's check if the node exists in the graph
        if node is not None and self._lineage_graph.has_node(node):
            node_successors = list(self._lineage_graph.successors(node))
            node_predecessors = list(self._lineage_graph.predecessors(node))

            # networknx's remove_node already takes care of in and out edges
            self._lineage_graph.remove_node(node)
            if node in self._graph_attributes:
                del self._graph_attributes[node]

            # Now that we have just deleted the dropped table from the graph, we need to take care of
            # new island nodes.
            if not self._show_isolated_nodes:
                for successor in node_successors:
                    if self._lineage_graph.has_node(successor) and self._lineage_graph.degree(successor) == 0:
                        self._lineage_graph.remove_node(successor)
                for predecessor in node_predecessors:
                    if self._lineage_graph.has_node(predecessor) and self._lineage_graph.degree(predecessor) == 0:
                        self._lineage_graph.remove_node(predecessor)

    def init_graph_from_query_list(self, queries: [Query]) -> None:
        queries_with_progress_bar = alive_it(queries, title='Updating lineage graph')
        for query in queries_with_progress_bar:
            self._update_lineage_graph(query)

        logger.debug(f'Finished updating lineage graph!')

    @staticmethod
    def parse_filter(filter_str: str) -> (Union[int, None], str, Union[int, None]):
        split_filter = [split_str.strip() for split_str in filter_str.split('+')]
        upstream_depth = 0
        downstream_depth = 0
        selected_item = None

        if len(split_filter) == 1:
            selected_item = split_filter[0]
        elif len(split_filter) == 2:
            if split_filter[0].isnumeric():
                upstream_depth = int(split_filter[0])
                selected_item = split_filter[1]
            elif split_filter[1].isnumeric():
                selected_item = split_filter[0]
                downstream_depth = int(split_filter[1])
            elif split_filter[0] == '':
                upstream_depth = None
                selected_item = split_filter[1]
            elif split_filter[1] == '':
                selected_item = split_filter[0]
                downstream_depth = None
        elif len(split_filter) == 3:
            if split_filter[0].isnumeric():
                upstream_depth = int(split_filter[0])
            elif split_filter[0] == '':
                upstream_depth = None
            if split_filter[2].isnumeric():
                downstream_depth = int(split_filter[2])
            elif split_filter[2] == '':
                downstream_depth = None
            selected_item = split_filter[1]

        if selected_item is None:
            raise ConfigError('Invalid graph filter')

        return upstream_depth, selected_item, downstream_depth

    def filter(self, database_filter: str, schema_filter: str, table_filter: str) -> None:
        if table_filter is not None:
            self.filter_on_table(table_filter)
            return
        if schema_filter is not None:
            self.filter_on_schema(schema_filter)
            return
        if database_filter is not None:
            self.filter_on_database(database_filter)
            return

    def get_subgraph(self, nodes: set, upstream_depth: Union[int, None], downstream_depth: Union[int, None]) \
            -> nx.DiGraph:
        subgraph = nx.DiGraph()
        subgraph.add_nodes_from((n, self._lineage_graph.nodes[n]) for n in nodes)
        subgraph.add_edges_from((n, nbr, d) for n, nbrs in self._lineage_graph.adj.items() if n in nodes
                                for nbr, d in nbrs.items() if nbr in nodes)

        if upstream_depth is None or upstream_depth > 0:
            source_nodes = [x for x in subgraph.nodes() if subgraph.in_degree(x) == 0]
            for node in source_nodes:
                subgraph = nx.compose(subgraph, self._upstream_graph(node, upstream_depth))
        if downstream_depth is None or downstream_depth > 0:
            last_nodes = [x for x in subgraph.nodes() if subgraph.out_degree(x) == 0]
            for node in last_nodes:
                subgraph = nx.compose(subgraph, self._downstream_graph(node, downstream_depth))

        return subgraph

    def filter_on_database(self, database_filter: str) -> None:
        upstream_depth, database_name, downstream_depth = self.parse_filter(database_filter)
        nodes_in_db = set()
        for node in self._lineage_graph:
            if database_name.lower() == node.split('.')[0].lower():
                nodes_in_db.add(node)

        self._lineage_graph = self.get_subgraph(nodes_in_db, upstream_depth, downstream_depth)

    def filter_on_schema(self, schema_filter: str) -> None:
        upstream_depth, schema_name, downstream_depth = self.parse_filter(schema_filter)
        nodes_in_schema = set()
        for node in self._lineage_graph:
            node_database_name, node_schema_name, node_table_name = [part.lower() for part in node.split('.')]
            normalized_schema_name = schema_name.lower()
            if normalized_schema_name == node_schema_name or normalized_schema_name == \
                    '.'.join([node_database_name, node_schema_name]):
                nodes_in_schema.add(node)

        self._lineage_graph = self.get_subgraph(nodes_in_schema, upstream_depth, downstream_depth)

    def filter_on_table(self, table_filter: str) -> None:
        upstream_depth, table_name, downstream_depth = self.parse_filter(table_filter)
        matched_nodes = set()
        for node in self._lineage_graph.nodes:
            node_database_name, node_schema_name, node_table_name = [part.lower() for part in node.split('.')]
            normalized_table_name = table_name.lower()
            if normalized_table_name == node_table_name or normalized_table_name == '.'.join([node_schema_name,
                                                                                              node_table_name]):
                matched_nodes.add(node)

        self._lineage_graph = self.get_subgraph(matched_nodes, upstream_depth, downstream_depth)
        if len(matched_nodes) == 1:
            self._update_selected_node_attributes(matched_nodes.pop())

    def _downstream_graph(self, source_node: str, depth: Optional[int]) -> nx.DiGraph:
        logger.debug(f'Building a downstream graph for - {source_node}, depth - {depth}')
        return nx.bfs_tree(G=self._lineage_graph, source=source_node, depth_limit=depth)

    def _upstream_graph(self, target_node: str, depth: Optional[int]) -> nx.DiGraph:
        logger.debug(f'Building an upstream graph for - {target_node}, depth - {depth}')
        reversed_lineage_graph = self._lineage_graph.reverse(copy=True)
        return nx.bfs_tree(G=reversed_lineage_graph, source=target_node, depth_limit=depth).reverse(copy=False)

    def _update_selected_node_attributes(self, selected_node: str) -> None:
        if self._lineage_graph.has_node(selected_node):
            node = self._lineage_graph.nodes[selected_node]
            node_title = node.get('title', '')
            node.update({'color': self.SELECTED_NODE_COLOR,
                         'title': self.SELECTED_NODE_TITLE + node_title})

    @staticmethod
    def _load_header() -> str:
        header_content = ""
        header = pkg_resources.resource_filename(__name__, "header.html")

        if header is not None:
            with open(header, 'r') as header_file:
                header_content = header_file.read()

        return header_content

    def properties(self):
        return {'lineage_properties': {'nodes_count': len(self._lineage_graph.nodes),
                                       'edges_count': len(self._lineage_graph.edges)}}

    def _enrich_graph_with_monitoring_data(self):
        for node, attr in self._lineage_graph.nodes(data=True):
            if node in self._graph_attributes:
                self._lineage_graph.nodes[node]['title'] = attr.get('title', '') + self._graph_attributes[node]

    def export_graph_to_files(self, target_dir_path: str) -> None:
        lineage_graph_file_path = os.path.join(target_dir_path, self.LINEAGE_GRAPH_FILE_MAME)
        lineage_graph_attributes_file_path = os.path.join(target_dir_path,
                                                          self.LINEAGE_GRAPH_ATTRIBUTES_FILE_NAME)

        nx.write_gpickle(self._lineage_graph, lineage_graph_file_path)
        with open(lineage_graph_attributes_file_path, 'w') as graph_attributes_file:
            json.dump(self._graph_attributes, graph_attributes_file)

    def load_graph_from_files(self, target_dir_path: str) -> bool:
        lineage_graph_file_path = os.path.join(target_dir_path, self.LINEAGE_GRAPH_FILE_MAME)
        lineage_graph_attributes_file_path = os.path.join(target_dir_path,
                                                          self.LINEAGE_GRAPH_ATTRIBUTES_FILE_NAME)

        if not os.path.exists(lineage_graph_file_path) or not os.path.exists(lineage_graph_attributes_file_path):
            return False

        self._lineage_graph = nx.read_gpickle(lineage_graph_file_path)
        with open(lineage_graph_attributes_file_path, 'r') as graph_attributes_file:
            self._graph_attributes = json.load(graph_attributes_file)

        return True

    def draw_graph(self, should_open_browser: bool = True, full_table_names: bool = True) -> bool:
        if len(self._lineage_graph.edges) == 0 and len(self._lineage_graph.nodes) == 0:
            return False

        self._enrich_graph_with_monitoring_data()

        if not full_table_names:
            short_name_mappings = dict()
            for node in self._lineage_graph.nodes:
                short_name_mappings[node] = node.rsplit('.', 1)[-1]
            self._lineage_graph = nx.relabel_nodes(self._lineage_graph, short_name_mappings)

        # Visualize the graph
        net = Network(height="95%", width="100%", directed=True, heading=self._load_header())
        net.from_nx(self._lineage_graph)
        net.set_options(GRAPH_VISUALIZATION_OPTIONS)

        net.save_graph("elementary_lineage.html")
        if should_open_browser:
            webbrowser.open_new_tab('elementary_lineage.html')

        return True




import alive_progress
import time
from threading import Thread


class ThreadSpinner(object):
    def __init__(self, title: str) -> None:
        self.stop_spinner = False
        self.thread = Thread(target=self._run_spinner, args=(title,))

    def _run_spinner(self, title: str) -> None:
        with alive_progress.alive_bar(title=title, bar=None, spinner='twirl', monitor=False) as spinner:
            while not self.stop_spinner:
                time.sleep(0.2)
                spinner()

    def start(self):
        self.thread.start()

    def stop(self):
        self.stop_spinner = True
        self.thread.join()
        self.stop_spinner = False

    def __enter__(self):
        self.start()

    def __exit__(self, type, value, traceback):
        self.stop()



import os
import os.path
from pathlib import Path
from typing import Dict, Any, Union
import dbt.config
from dbt.context.base import generate_base_context
from dbt.exceptions import DbtConfigError
from dbt.adapters.bigquery.connections import BigQueryConnectionManager
import google.cloud.bigquery
import google.cloud.exceptions
from google.api_core import client_info
from exceptions.exceptions import ConfigError
from utils.log import get_logger
import snowflake.connector

from utils.ordered_yaml import OrderedYaml

logger = get_logger(__name__)

DBT_DEFAULT_DIR = ".dbt"


def extract_profile_data(profiles_raw: Dict[str, Any], profile_name: str, target_name: str) -> Dict[str, Any]:
    profile_data = dict()
    try:
        selected_profile = profiles_raw[profile_name]
        profile_data = selected_profile['outputs'][target_name]
    except KeyError as exc:
        logger.debug(f"Failed extracting profile data: {profiles_raw}, {profile_name}, {target_name}, {exc}")

    return profile_data


def extract_credentials_and_data_from_profiles(profiles_dir: str, profile_name: str):
    try:
        profiles_raw = dbt.config.profile.read_profile(profiles_dir)
        empty_profile_renderer = dbt.config.renderer.ProfileRenderer(generate_base_context({}))
        dbt_profile = dbt.config.Profile.from_raw_profiles(profiles_raw, profile_name, empty_profile_renderer)
        profile_data = extract_profile_data(profiles_raw, profile_name, dbt_profile.target_name)
        return dbt_profile.credentials, profile_data
    except DbtConfigError as exc:
        logger.debug(f"Failed parsing selected profile - {profiles_dir}, {profile_name}, {exc}")
        raise ConfigError(f"Failed parsing selected profile - {profiles_dir}, {profile_name}")


def get_bigquery_client(profile_credentials):
    if profile_credentials.impersonate_service_account:
        creds = \
            BigQueryConnectionManager.get_impersonated_bigquery_credentials(profile_credentials)
    else:
        creds = BigQueryConnectionManager.get_bigquery_credentials(profile_credentials)

    database = profile_credentials.database
    location = getattr(profile_credentials, 'location', None)

    info = client_info.ClientInfo(user_agent=f'elementary')
    return google.cloud.bigquery.Client(
        database,
        creds,
        location=location,
        client_info=info,
    )


def get_snowflake_client(profile_credentials, server_side_binding=True):
    if server_side_binding:
        snowflake.connector.paramstyle = 'numeric'

    return snowflake.connector.connect(
        account=profile_credentials.account,
        user=profile_credentials.user,
        database=profile_credentials.database,
        schema=profile_credentials.schema,
        warehouse=profile_credentials.warehouse,
        role=profile_credentials.role,
        autocommit=True,
        client_session_keep_alive=profile_credentials.client_session_keep_alive,
        application='elementary',
        **profile_credentials.auth_args()
    )


def get_profile_name_from_dbt_project(dbt_project_path: str) -> str:
    ordered_yaml = OrderedYaml()
    dbt_project_dict = ordered_yaml.load(os.path.join(dbt_project_path, 'dbt_project.yml'))
    return dbt_project_dict['profile']


def get_model_paths_from_dbt_project(dbt_project_path: str) -> list:
    ordered_yaml = OrderedYaml()
    dbt_project_dict = ordered_yaml.load(os.path.join(dbt_project_path, 'dbt_project.yml'))
    return dbt_project_dict.get('model-paths', dbt_project_dict.get('source-paths', ['models']))


def get_target_database_name(profiles_dir: str, dbt_project_path: str) -> Union[str, None]:
    try:
        profile_name = get_profile_name_from_dbt_project(dbt_project_path)
        credentials, profile_data = extract_credentials_and_data_from_profiles(profiles_dir, profile_name)
        return credentials.database
    except Exception:
        pass
    return None


def is_dbt_installed() -> bool:
    if os.path.exists(os.path.join(Path.home(), DBT_DEFAULT_DIR)):
        return True
    return False



from collections import defaultdict
from lineage.query_context import QueryContext


class QueryHistoryStats(object):
    def __init__(self):
        self._query_type_stats = defaultdict(lambda: 0)
        self._roles = set()
        self._users = set()

    def update_stats(self, query_context: QueryContext):
        if query_context.query_type is not None:
            self._query_type_stats[query_context.query_type] += 1

        if query_context.user_name is not None:
            self._users.add(query_context.user_name)

        if query_context.role_name is not None:
            self._roles.add(query_context.role_name)

    def to_dict(self):
        query_history_stats_dict = self._query_type_stats.copy()
        query_history_stats_dict['user_count'] = len(self._users)
        query_history_stats_dict['role_count'] = len(self._roles)
        return query_history_stats_dict



import click
import os
from os.path import expanduser

from config.config import Config
from exceptions.exceptions import ConfigError
from lineage.query_history_factory import QueryHistoryFactory
from utils.package import get_package_version
from utils.dbt import is_dbt_installed
from lineage.empty_graph_helper import EmptyGraphHelper
from tracking.anonymous_tracking import track_cli_start, track_cli_end, track_cli_exception, AnonymousTracking
from datetime import timedelta, date
from lineage.lineage_graph import LineageGraph
from datetime import datetime


class RequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_if = kwargs.pop('required_if')
        assert self.required_if, "'required_if' parameter required"
        kwargs['help'] = (kwargs.get('help', '') +
                          ' NOTE: This argument must be configured together with %s.' %
                          self.required_if
                          ).strip()
        super(RequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.required_if in opts

        if we_are_present and not other_present:
            raise click.UsageError(
                "Illegal usage: `%s` must be configured with `%s`" % (
                    self.name, self.required_if))
        else:
            self.prompt = None

        return super(RequiredIf, self).handle_parse_result(
            ctx, opts, args)


def get_cli_lineage_properties() -> dict:

    click_context = click.get_current_context()
    if click_context is None:
        return dict()

    params = click_context.params
    if params is None:
        return dict()

    is_filtered = False
    if params.get('table') is not None or params.get('schema') is not None or params.get('database') is not None:
        is_filtered = True

    return {'is_filtered': is_filtered,
            'open_browser': params.get('open_browser'),
            'full_table_names': params.get('full_table_names'),
            'command': click_context.command.name,
            'dbt_installed': is_dbt_installed(),
            'version': get_package_version()}


def get_cli_lineage_generate_properties() -> dict:

    click_context = click.get_current_context()
    if click_context is None:
        return dict()

    params = click_context.params
    if params is None:
        return dict()

    start_date = params.get('start_date')
    end_date = params.get('end_date')
    limited_dbs = True if params.get('databases') is not None else False

    start_date_str = None
    if start_date is not None:
        start_date_str = start_date.isoformat()

    end_date_str = None
    if end_date is not None:
        end_date_str = end_date.isoformat()

    return {'start_date': start_date_str,
            'end_date': end_date_str,
            'limited_dbs': limited_dbs,
            'dbt_installed': is_dbt_installed(),
            'version': get_package_version()}


@click.group(invoke_without_command=True)
@click.option(
    '--database', '-db',
    type=str,
    help="Filter on a database to see upstream and downstream dependencies of this database "
         "(use X+<database_name>+Y to see X upstream dependencies and Y downstream dependencies, "
         "operator '+' could also be used without a depth limit).",
    default=None
)
@click.option(
    '--schema', '-sch',
    type=str,
    default=None,
    help="Filter on a schema to see upstream and downstream dependencies of this schema "
         "(use X+<schema_name>+Y to see X upstream dependencies and Y downstream dependencies, "
         "operator '+' could also be used without a depth limit).",
)
@click.option(
    '--table', '-t',
    type=str,
    help="Filter on a table to see upstream and downstream dependencies of this table "
         "(use X+<table_name>+Y to see X upstream dependencies and Y downstream dependencies, "
         "operator '+' could also be used without a depth limit).",
    default=None
)
@click.option(
    '--open-browser', '-o',
    type=bool,
    default=True,
    help="Indicates if the data lineage graph should be opened in your default browser, if this flag is set to "
         "no-browser an html file will be saved to your current directory instead of opening it in your browser "
         "(by default the lineage will be opened automatically in your browser).",
)
@click.option(
    '--full-table-names', '-n',
    type=bool,
    default=True,
    help="Indicates if the lineage should display full table names including the relevant database and schema names "
         "(the default is to show full table names)."
)
@click.option(
    '--config-dir', '-c',
    type=str,
    default=os.path.join(expanduser('~'), '.edr'),
    help="Global settings for edr are configured in a config.yml file in this directory "
         "(if your config dir is HOME_DIR/.edr, no need to provide this parameter as we use it as default)."
)
@click.option(
    '--profiles-dir', '-d',
    type=click.Path(exists=True),
    default=os.path.join(expanduser('~'), '.dbt'),
    help="Specify your profiles dir where a profiles.yml is located, this could be a dbt profiles dir "
         "(if your profiles dir is HOME_DIR/.dbt, no need to provide this parameter as we use it as default).",
    cls=RequiredIf,
    required_if='profile_name'
)
@click.option(
    '--profile-name', '-p',
    type=str,
    default='elementary',
    help="The profile in your profiles.yml file that will be used as connection details to your data warehouse "
         "(if you configured 'elementary' profile in your profiles.yml, no need to provide this parameter as we use it "
         "as default).",
    cls=RequiredIf,
    required_if='profiles_dir'
)
@click.pass_context
def lineage(ctx, database: str, schema: str, table: str, open_browser: bool, full_table_names: bool,
            config_dir: str, profiles_dir: str, profile_name: str) -> None:
    click.echo(f"Any feedback and suggestions are welcomed! join our community here - "
               f"https://bit.ly/slack-elementary\n")
    if ctx.invoked_subcommand is not None:
        return

    config = Config(config_dir, profiles_dir, profile_name)
    anonymous_tracking = AnonymousTracking(config)
    track_cli_start(anonymous_tracking, 'lineage', get_cli_lineage_properties(), ctx.command.name)
    execution_properties = dict()

    try:
        lineage_graph = LineageGraph()
        success = lineage_graph.load_graph_from_files(config.target_dir)
        if not success:
            dbs = input("Lineage files were not found, please provide all database names that should be included: ")
            ctx.invoke(generate, databases=dbs)
            success = lineage_graph.load_graph_from_files(config.target_dir)
            if not success:
                raise ConfigError("Could not find lineage files, please run 'edr lineage generate' first")
            execution_properties['lineage_files_were_generated'] = True

        lineage_graph.filter(database, schema, table)
        success = lineage_graph.draw_graph(should_open_browser=open_browser, full_table_names=full_table_names)
        if not success:
            print(EmptyGraphHelper.get_help_message())

        execution_properties.update(lineage_graph.properties())
        track_cli_end(anonymous_tracking, 'lineage', execution_properties, ctx.command.name)
    except Exception as exc:
        track_cli_exception(anonymous_tracking, 'lineage', exc, ctx.command.name)
        raise


@lineage.command()
@click.option(
    '--start-date', '-s',
    type=click.DateTime(formats=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]),
    default=str(date.today() - timedelta(days=6)),
    help="Parse queries in query log since this start date (you could also provide a specific time), "
         "default start date is 7 days ago."
)
@click.option(
    '--end-date', '-e',
    type=click.DateTime(formats=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]),
    default=None,
    help="Parse queries in query log up to this end date (end_date is inclusive, "
         "you could also provide a specific time), default is current time."
)
@click.option(
    '--databases', '-dbs',
    type=str,
    default=None,
    required=True,
    help="Generate the lineage graph for databases that are included in this list "
         "(to provide more than one database name use ',' between the db names - <db_name1>,<db_name2>).."
)
@click.option(
    '--config-dir', '-c',
    type=str,
    default=os.path.join(expanduser('~'), '.edr'),
    help="Global settings for edr are configured in a config.yml file in this directory "
         "(if your config dir is HOME_DIR/.edr, no need to provide this parameter as we use it as default)."
)
@click.option(
    '--profiles-dir', '-d',
    type=click.Path(exists=True),
    default=os.path.join(expanduser('~'), '.dbt'),
    help="Specify your profiles dir where a profiles.yml is located, this could be a dbt profiles dir "
         "(if your profiles dir is HOME_DIR/.dbt, no need to provide this parameter as we use it as default).",
    cls=RequiredIf,
    required_if='profile_name'
)
@click.option(
    '--profile-name', '-p',
    type=str,
    default='elementary',
    help="The profile in your profiles.yml file that will be used as connection details to your data warehouse "
         "(if you configured 'elementary' profile in your profiles.yml, no need to provide this parameter as we use it "
         "as default).",
    cls=RequiredIf,
    required_if='profiles_dir'
)
@click.pass_context
def generate(ctx, start_date: datetime, end_date: datetime, databases: str, config_dir: str,
             profiles_dir: str, profile_name: str):
    config = Config(config_dir, profiles_dir, profile_name)
    anonymous_tracking = AnonymousTracking(config)
    track_cli_start(anonymous_tracking, 'lineage', get_cli_lineage_generate_properties(), ctx.command.name)

    try:
        query_history_extractor = QueryHistoryFactory.create_query_history(config, databases)
        queries = query_history_extractor.extract_queries(start_date, end_date)

        lineage_graph = LineageGraph()
        lineage_graph.init_graph_from_query_list(queries)
        lineage_graph.export_graph_to_files(config.target_dir)

        execution_properties = query_history_extractor.properties()
        execution_properties.update(lineage_graph.properties())

        track_cli_end(anonymous_tracking, 'lineage', execution_properties, ctx.command.name)
    except Exception as exc:
        track_cli_exception(anonymous_tracking, 'lineage', exc, ctx.command.name)
        raise


if __name__ == "__main__":
    lineage()



import os
import csv
from typing import Union
import glob

from exceptions.exceptions import ConfigError
from utils.dbt import get_model_paths_from_dbt_project, get_target_database_name, \
    extract_credentials_and_data_from_profiles
from utils.ordered_yaml import OrderedYaml

ordered_yaml = OrderedYaml()


class Config(object):
    SLACK_NOTIFICATION_WEBHOOK = 'slack_notification_webhook'
    CONFIG_FILE_NAME = 'config.yml'

    def __init__(self, config_dir: str, profiles_dir: str, profile_name: str) -> None:
        self.config_dir = config_dir
        self.profiles_dir = profiles_dir
        self.profile_name = profile_name
        self.credentials, self.profiles_data = extract_credentials_and_data_from_profiles(profiles_dir,
                                                                                          profile_name)
        self.config_dict = self._load_configuration()

    def _load_configuration(self) -> dict:
        if not os.path.exists(self.config_dir):
            os.makedirs(self.config_dir)

        config_file_path = os.path.join(self.config_dir, self.CONFIG_FILE_NAME)
        if not os.path.exists(config_file_path):
            return {}

        return ordered_yaml.load(config_file_path)

    @property
    def query_history_source(self):
        return self.profiles_data.get('query_history_source')

    @property
    def platform(self):
        return self.profiles_data.get('type', 'unknown')

    @property
    def anonymous_tracking_enabled(self) -> bool:
        return self.config_dict.get('anonymous_usage_tracking', True)

    @property
    def slack_notification_webhook(self) -> Union[str, None]:
        return self.config_dict.get(self.SLACK_NOTIFICATION_WEBHOOK)

    @property
    def target_dir(self) -> str:
        target_path = self.config_dict.get('target-path')
        if not target_path:
            return os.getcwd()
        return target_path

    @staticmethod
    def _find_schema_yml_files_in_dbt_project(dbt_project_models_path: str) -> list:
        return glob.glob(os.path.join(dbt_project_models_path, '*.yml'), recursive=True)

    def _get_sources_from_all_dbt_projects(self) -> list:
        config_dict = self._load_configuration()
        dbt_projects = config_dict.get('dbt_projects', [])
        sources = []
        for dbt_project_path in dbt_projects:
            try:
                dbt_project_target_database = get_target_database_name(self.profiles_dir, dbt_project_path)
                model_paths = get_model_paths_from_dbt_project(dbt_project_path)
                for model_path in model_paths:
                    dbt_project_models_path = os.path.join(dbt_project_path, model_path)
                    schema_yml_files = self._find_schema_yml_files_in_dbt_project(dbt_project_models_path)
                    for schema_yml_file in schema_yml_files:
                        schema_dict = ordered_yaml.load(schema_yml_file)
                        schema_sources = schema_dict.get('sources')
                        if schema_sources is not None:
                            dbt_project_sources = {'sources': schema_sources,
                                                   'dbt_project_target_database': dbt_project_target_database}
                            sources.append(dbt_project_sources)
            except FileNotFoundError as exc:
                raise ConfigError(f'No such file - {exc.filename}, please configure a valid dbt project path')
        return sources

    @staticmethod
    def _alert_on_schema_changes(source_dict: dict) -> Union[bool, None]:
        metadata = source_dict.get('meta', {})
        edr_config = metadata.get('edr', {})
        alert_on_schema_changes = edr_config.get('schema_changes')

        # Normalize alert_on_schema_changes to handle both booleans and strings
        alert_on_schema_changes_str = str(alert_on_schema_changes).lower()
        if alert_on_schema_changes_str == 'false':
            return False
        elif alert_on_schema_changes_str == 'true':
            return True
        else:
            return None

    def monitoring_configuration_in_dbt_sources_to_csv(self, target_csv_path: str) -> int:
        row_count = 0
        with open(target_csv_path, 'w') as target_csv:
            target_csv_writer = csv.DictWriter(target_csv, fieldnames=['database_name',
                                                                       'schema_name',
                                                                       'table_name',
                                                                       'column_name',
                                                                       'alert_on_schema_changes'])
            target_csv_writer.writeheader()

            all_configured_sources = self._get_sources_from_all_dbt_projects()
            for sources_dict in all_configured_sources:
                sources = sources_dict.get('sources', [])
                target_database = sources_dict.get('dbt_project_target_database')
                for source in sources:
                    source_db = source.get('database', target_database)
                    if source_db is None:
                        continue

                    schema_name = source.get('schema', source.get('name'))
                    if schema_name is None:
                        continue

                    alert_on_schema_changes = self._alert_on_schema_changes(source)
                    if alert_on_schema_changes is not None:
                        target_csv_writer.writerow({'database_name': source_db,
                                                    'schema_name': schema_name,
                                                    'table_name': None,
                                                    'column_name': None,
                                                    'alert_on_schema_changes': alert_on_schema_changes})
                        row_count += 1

                    source_tables = source.get('tables', [])
                    for source_table in source_tables:
                        table_name = source_table.get('identifier', source_table.get('name'))
                        if table_name is None:
                            continue

                        alert_on_schema_changes = self._alert_on_schema_changes(source_table)
                        if alert_on_schema_changes is not None:
                            target_csv_writer.writerow({'database_name': source_db,
                                                        'schema_name': schema_name,
                                                        'table_name': table_name,
                                                        'column_name': None,
                                                        'alert_on_schema_changes': alert_on_schema_changes})
                            row_count += 1

                        source_columns = source_table.get('columns', [])
                        for source_column in source_columns:
                            column_name = source_column.get('name')
                            if column_name is None:
                                continue

                            alert_on_schema_changes = self._alert_on_schema_changes(source_column)
                            if alert_on_schema_changes is not None:
                                target_csv_writer.writerow({'database_name': source_db,
                                                            'schema_name': schema_name,
                                                            'table_name': table_name,
                                                            'column_name': column_name,
                                                            'alert_on_schema_changes': alert_on_schema_changes})
                                row_count += 1
        return row_count



from typing import Union

from lineage.query_context import QueryContext
from lineage.table_resolver import TableResolver
from lineage.query import Query
from utils.log import get_logger
import re
import sqlfluff

logger = get_logger(__name__)

DOLLAR_SIGN_REGEX = re.compile(r'(\b\S+)(\$)(\S+\b)')
DOLLAR_SIGN_PLACEHOLDER = '__dollar_sign__'


class SnowflakeQuery(Query):
    PLATFORM_TYPE = 'SNOWFLAKE'

    DDL_QUERY_TYPES_PREFIX = {'DROP', 'RENAME', 'ALTER'}

    @staticmethod
    def from_dict(query_dict: dict):
        query_context = QueryContext.from_dict(query_dict.pop('query_context'))
        if 'platform_type' in query_dict:
            query_dict.pop('platform_type')
        return SnowflakeQuery(**query_dict, query_context=query_context)

    @staticmethod
    def replace_dollar_sign_with_placeholder(query_text: str) -> str:
        return re.sub(DOLLAR_SIGN_REGEX, f'\\1{DOLLAR_SIGN_PLACEHOLDER}\\3', query_text)

    @staticmethod
    def revert_dollar_sign_placeholder(name: str) -> str:
        return name.replace(DOLLAR_SIGN_PLACEHOLDER, '$')

    @staticmethod
    def _parse_merge_query(table_resolver: TableResolver, raw_query_text: str) -> (set, set):
        source_tables = set()
        target_tables = set()
        parsed_query = sqlfluff.parse(raw_query_text, dialect='snowflake')
        merge_stmts = list(parsed_query.tree.recursive_crawl('merge_statement'))
        if len(merge_stmts) != 1:
            return source_tables, target_tables
        merge_stmt = merge_stmts[0]
        look_for_target = False
        look_for_source = False
        for seg in merge_stmt.segments:
            if seg.is_type('keyword'):
                if seg.raw.lower() == 'into':
                    look_for_target = True
                    look_for_source = False
                elif seg.raw.lower() == 'using':
                    look_for_source = True
                    look_for_target = False
                else:
                    look_for_target = False
                    look_for_source = False
                continue

            if seg.is_type('table_reference') and look_for_target:
                target_tables.add(table_resolver.name_qualification(seg.raw))

            if look_for_source:
                source_table_references = list(seg.recursive_crawl('table_reference'))
                for source_table_reference in source_table_references:
                    source_tables.add(table_resolver.name_qualification(source_table_reference.raw))

        return source_tables, target_tables

    def _parse_platform_specific_queries(self, table_resolver: TableResolver, raw_query_text: str) -> (set, set):
        source_tables = set()
        target_tables = set()
        try:
            query_type = self.query_context.query_type
            if query_type is not None and query_type.lower() == 'merge':
                source_tables, target_tables = self._parse_merge_query(table_resolver, raw_query_text)
        except Exception as exc:
            logger.debug(f'Exception was raised while parsing this query with sqlfluff -\n{raw_query_text}\n'
                         f'Error was -\n{exc}.')
        return source_tables, target_tables

    @classmethod
    def _is_ddl(cls, query_type: Union[str, None]):
        if query_type is None:
            return False

        for ddl_query_type_prefix in cls.DDL_QUERY_TYPES_PREFIX:
            if query_type.startswith(ddl_query_type_prefix):
                return True

        return False

    def parse(self, full_table_names: bool = False) -> bool:
        try:
            table_resolver = TableResolver(self.query_context.queried_database, self.query_context.queried_schema,
                                           full_table_names, self.revert_dollar_sign_placeholder)

            ddl_query = self._is_ddl(self.query_context.query_type)
            if self.query_context.destination_table is not None and self.query_context.referenced_tables is not \
                    None and not ddl_query:
                self.source_tables = {table_resolver.name_qualification(source_table) for source_table in
                                      self.query_context.referenced_tables}
                self.target_tables = {table_resolver.name_qualification(self.query_context.destination_table)}
            else:
                # sqlparse library doesn't behave nicely when there is a $ sign in the table name.
                # Therefore we replace it with a placeholder (and revert it back later on using our table resolver)
                raw_query_text = self.replace_dollar_sign_with_placeholder(self._raw_query_text)
                self.source_tables, self.target_tables, self.renamed_tables, self.dropped_tables = \
                    self._parse_query_text(table_resolver, raw_query_text)

            return True

        except Exception as exc:
            logger.debug(f'Exception was raised while parsing this query -\n{self._raw_query_text}\n'
                         f'Error was -\n{exc}.')
        return False




import logging
import sys

from utils.env_vars import is_debug_mode_on

FORMATTER = logging.Formatter("%(asctime)s — %(name)s — %(levelname)s — %(message)s")
LOG_FILE = "edr.log"


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = logging.FileHandler(LOG_FILE)
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    log_level = logging.DEBUG if is_debug_mode_on() else logging.INFO
    logger.setLevel(log_level)
    logger.addHandler(get_console_handler())
    logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to parent
    logger.propagate = False
    return logger



class EmptyGraphHelper(object):
    @staticmethod
    def get_help_message():
        return f"""
        We are deeply sorry but unfortunately the result graph is empty.
        Please try the following steps to fix the problem - 
        \t1. Try running again with other / no filters
        \t2. Make sure to run 'edr lineage generate' command first
        \t3. Join our slack channel here - https://bit.ly/slack-elementary, we promise to help and be nice! 
        """



from typing import Optional

import pkg_resources


def get_package_version() -> Optional[str]:
    try:
        return pkg_resources.get_distribution('elementary-data').version
    except Exception:
        pass

    return None


from sqllineage.models import Schema, Table
from typing import Optional, Union, Callable
from utils.log import get_logger

logger = get_logger(__name__)


class TableResolver(object):

    def __init__(self, database_name: str = None, schema_name: str = None, full_table_names: bool = False,
                 remove_special_char_callback: Callable = None) -> None:
        self._database_name = database_name
        self._schema_name = schema_name
        self._show_full_table_name = full_table_names
        self._remove_special_char_callback = remove_special_char_callback

    @staticmethod
    def _resolve_table_qualification(table: Table, database_name: str, schema_name: str) -> Table:
        if not table.schema:
            if database_name is not None and schema_name is not None:
                table.schema = Schema(f'{database_name}.{schema_name}')
        else:
            parsed_query_schema_name = str(table.schema)
            if '.' not in parsed_query_schema_name:
                # Resolved schema is either empty or fully qualified with db_name.schema_name
                if database_name is not None:
                    table.schema = Schema(f'{database_name}.{parsed_query_schema_name}')
                else:
                    table.schema = Schema()
        return table

    def name_qualification(self, table: Union[Table, str]) -> Optional[str]:
        if isinstance(table, str):
            table = Table(table)

        table = self._resolve_table_qualification(table, self._database_name, self._schema_name)

        if self._show_full_table_name:
            resolved_table_name = str(table)
        else:
            resolved_table_name = str(table).rsplit('.', 1)[-1]

        if self._remove_special_char_callback is not None:
            resolved_table_name = self._remove_special_char_callback(resolved_table_name)

        logger.debug(f'Resolved table name - {resolved_table_name}')
        return resolved_table_name



from datetime import datetime, date, timedelta
from typing import Union

from alive_progress import alive_it
from exceptions.exceptions import ConfigError
from lineage.query_context import QueryContext
from lineage.query_history import QueryHistory
from lineage.snowflake_query import SnowflakeQuery
from utils.log import get_logger
from utils.thread_spinner import ThreadSpinner

logger = get_logger(__name__)


class SnowflakeQueryHistory(QueryHistory):
    PLATFORM_TYPE = 'snowflake'

    # Note: Here we filter permissively on the configured database_name, basically finding all the queries that are
    # relevant to this database. Snowflake's query history might show in the database_name column a different db name
    # than the db name that was part of the query. During the parsing logic in the lineage graph we strictly analyze if
    # the query was really executed on the configured db and filter it accordingly.

    INFORMATION_SCHEMA_QUERY_HISTORY_AND_VIEWS = """
    {database_name_normalized}_query_history as (
        select 
            query_text, 
            database_name, 
            schema_name, 
            end_time,
            rows_produced, 
            query_type, 
            user_name, 
            role_name, 
            total_elapsed_time, 
            query_id,
            null as target_table,
            null as source_tables
      from table({database_name}.information_schema.query_history(end_time_range_start=>to_timestamp_ltz(%(start_date)s),
        {end_time_range_end_expr}))
        where execution_status = 'SUCCESS' and query_type not in 
        ('SHOW', 'COPY', 'COMMIT', 'DESCRIBE', 'ROLLBACK', 'CREATE_STREAM', 'DROP_STREAM', 'PUT_FILES', 'GET_FILES',
         'BEGIN_TRANSACTION', 'GRANT', 'ALTER_SESSION', 'USE', 'ALTER_NETWORK_POLICY', 'ALTER_ACCOUNT', 
         'ALTER_TABLE_DROP_CLUSTERING_KEY', 'ALTER_USER',  'CREATE_CUSTOMER_ACCOUNT', 'CREATE_NETWORK_POLICY', 
         'CREATE_ROLE', 'CREATE_USER', 'DESCRIBE_QUERY', 'DROP_NETWORK_POLICY', 'DROP_ROLE', 'DROP_USER', 'LIST_FILES',
         'REMOVE_FILES', 'REVOKE','UNKNOWN', 'DELETE', 'SELECT') and
        (query_text not ilike '%%.query_history%%') and
        (lower(query_text) like '%%{database_name}%%' 
         or lower(database_name) = '{database_name}')
    ),
    {database_name_normalized}_views as (
        select 
            view_definition as query_text,
            table_catalog as database_name,
            table_schema as schema_name,
            last_altered as end_time,
            null as rows_produced,
            'CREATE_VIEW' as query_type,
            table_owner as user_name,
            null as role_name,
            null as total_elapsed_time,
            null as query_id,
            null as target_table,
            null as source_tables
        from {database_name}.information_schema.views
        where lower(table_schema) != 'information_schema' 
              and view_definition is not NULL
              and (lower(view_definition) like '%%{database_name}%%' 
                   or lower(table_catalog) = '{database_name}')
    ),
    {database_name_normalized}_history_and_views as (
        select * from {database_name_normalized}_query_history union all select * from {database_name_normalized}_views
    ),
    """

    SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY_AND_VIEWS = "select * from " \
                                                             "{database_name_normalized}_history_and_views"

    UNION_ALL_DBS = """
    union_all_dbs as (
        {union_all_dbs}
    )
    select * from union_all_dbs
    order by end_time
    """

    INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP = 'end_time_range_end=>to_timestamp_ltz(current_timestamp())'
    INFO_SCHEMA_END_TIME_UP_TO_PARAMETER = 'end_time_range_end=>to_timestamp_ltz(%(end_date)s)'
    QUERY_HISTORY_SOURCE_INFORMATION_SCHEMA = 'information_schema'

    ACCOUNT_USAGE_QUERY_HISTORY = """
    with source_access_history as (
        select *
        from snowflake.account_usage.access_history
        where query_start_time >= %(start_date)s  
        qualify row_number() over (partition by query_id order by query_id) = 1
    ),
    
    source_query_history as (
        select *
        from snowflake.account_usage.query_history
        where end_time >= %(start_date)s 
        and {end_time_range_end_expr}
        and (lower(query_text) like any (%(database_names_in_like_statement)s) 
             or lower(database_name) in (%(database_names)s))
        qualify row_number() over (partition by query_id order by query_id) = 1
    
    ),
    
    account_usage_views as (
        select * 
        from snowflake.account_usage.views
        where (lower(view_definition) like any (%(database_names_in_like_statement)s)
        or lower(table_catalog) in (%(database_names)s))
        and deleted is null 
        and view_definition is not NULL
      
    ),
    
    access_history as (
         select
             src.query_id,
             lower(replace(direct.value:"objectName"::varchar, '"','')) as direct_access_table_name,
             lower(replace(modified.value:"objectName"::varchar, '"','')) as modified_table_name
         from source_access_history as src,
            lateral flatten(input => src.direct_objects_accessed) as direct,
            lateral flatten(input => src.objects_modified) as modified
         where modified_table_name is not null and direct_access_table_name != modified_table_name       
    ),
    
    query_history as (
        select
            query_id,
            database_name,
            schema_name,
            query_text,
            query_type,
            role_name,
            user_name,
            rows_inserted + rows_updated as rows_produced,
            end_time,
            total_elapsed_time
        from source_query_history
        where is_client_generated_statement = false
              and (lower(query_text) not ilike '%%.query_history%%')
              and execution_status = 'SUCCESS'
              and query_type not in
              ('SHOW', 'COMMIT', 'DESCRIBE', 'ROLLBACK', 'CREATE_STREAM', 'DROP_STREAM',
                'BEGIN_TRANSACTION', 'GRANT', 'ALTER_SESSION', 'USE', 'ALTER_NETWORK_POLICY', 'ALTER_ACCOUNT',
                'ALTER_TABLE_DROP_CLUSTERING_KEY', 'ALTER_USER',  'CREATE_CUSTOMER_ACCOUNT', 'CREATE_NETWORK_POLICY',
                'CREATE_ROLE', 'CREATE_USER', 'DESCRIBE_QUERY', 'DROP_NETWORK_POLICY', 'DROP_ROLE', 'DROP_USER', 
                'LIST_FILES', 'REMOVE_FILES', 'REVOKE')
     ),
     
     views as (
        select 
            view_definition as query_text,
            table_catalog as database_name,
            table_schema as schema_name,
            last_altered as end_time,
            null as rows_produced,
            'CREATE_VIEW' as query_type,
            null as user_name,
            table_owner as role_name,
            null as total_elapsed_time,
            null as query_id,
            table_name as target_table,
            null as source_tables
       from account_usage_views
       where lower(table_schema) != 'information_schema'
     ),
     
    access_and_query_history as (
        select
            qh.query_text,
            qh.database_name,
            qh.schema_name,
            qh.end_time,
            qh.rows_produced,
            qh.query_type,
            qh.user_name,
            qh.role_name,
            qh.total_elapsed_time,
            qh.query_id,
            ah.modified_table_name as target_table,
            array_agg(distinct ah.direct_access_table_name) as source_tables
        from query_history as qh
            left join access_history as ah
            on (qh.query_id = ah.query_id)
       group by 1,2,3,4,5,6,7,8,9,10,11
      ),
      
    final as (
        select * from access_and_query_history
        union all
        select * from views
    )
    
    select * from final
    order by end_time
    """

    ACCOUNT_USAGE_END_TIME_UP_TO_CURRENT_TIMESTAMP = 'end_time <= current_timestamp()'
    ACCOUNT_USAGE_END_TIME_UP_TO_PARAMETER = 'end_time <= %(end_date)s'
    QUERY_HISTORY_SOURCE_ACCOUNT_USAGE = 'account_usage'

    def __init__(self, con, dbs: Union[str, None], should_export_query_history: bool = True,
                 full_table_names: bool = True, query_history_source: str = None) -> None:
        self.query_history_source = query_history_source.strip().lower() if query_history_source is not None else None
        self.access_history_queries = 0
        super().__init__(con, dbs, should_export_query_history, full_table_names)

    @classmethod
    def _info_schema_query_history(cls, start_date: datetime, end_date: datetime, dbs: list) -> (str, dict):
        if start_date.date() <= date.today() - timedelta(days=7):
            raise ConfigError(f"Cannot retrieve data from more than 7 days ago when pulling history from "
                              f"{cls.QUERY_HISTORY_SOURCE_INFORMATION_SCHEMA}, "
                              f"use {cls.QUERY_HISTORY_SOURCE_ACCOUNT_USAGE} instead "
                              f"(see https://docs.elementary-data.com for more details).")

        logger.debug("Pulling snowflake query history from information schema")
        if end_date is None:
            end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP
            params = {'start_date': start_date}
        else:
            end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_PARAMETER
            params = {'start_date': start_date,
                      'end_date': cls._include_end_date(end_date)}

        query_text = 'with'
        for db in dbs:
            query_text += cls.INFORMATION_SCHEMA_QUERY_HISTORY_AND_VIEWS. \
                format(database_name_normalized=cls._normalize_database_name(db),
                       database_name=db.lower(),
                       end_time_range_end_expr=end_time_range_end_expr)

        dbs_count = len(dbs)
        union_all_dbs = cls.SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY_AND_VIEWS. \
            format(database_name_normalized=cls._normalize_database_name(dbs[0]))
        for i in range(1, dbs_count):
            union_all_dbs += ' union all ' + cls.SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY_AND_VIEWS. \
                format(database_name_normalized=cls._normalize_database_name(dbs[i]))

        query_text += cls.UNION_ALL_DBS.format(union_all_dbs=union_all_dbs)
        return query_text, params

    @classmethod
    def _account_usage_query_history(cls, start_date: datetime, end_date: datetime, dbs: list) -> (str, dict):
        logger.debug("Pulling snowflake query history from account usage")
        if end_date is None:
            query_text = cls.ACCOUNT_USAGE_QUERY_HISTORY.format(end_time_range_end_expr=
                                                                cls.ACCOUNT_USAGE_END_TIME_UP_TO_CURRENT_TIMESTAMP)
            params = {'start_date': start_date,
                      'database_names': [db.lower() for db in dbs],
                      'database_names_in_like_statement': [f'%{db.lower()}%' for db in dbs]}
        else:
            query_text = cls.ACCOUNT_USAGE_QUERY_HISTORY.format(end_time_range_end_expr=
                                                                cls.ACCOUNT_USAGE_END_TIME_UP_TO_PARAMETER)
            params = {'start_date': start_date,
                      'end_date': cls._include_end_date(end_date),
                      'database_names': [db.lower() for db in dbs],
                      'database_names_in_like_statement': [f'%{db.lower()}%' for db in dbs]}

        return query_text, params

    def _query_history_table(self, start_date: datetime, end_date: datetime) -> None:
        logger.debug(f"Pulling snowflake query history for the following databases - {self._dbs}")

        with self._con.cursor() as cursor:
            if self.query_history_source == self.QUERY_HISTORY_SOURCE_ACCOUNT_USAGE:
                query_text, params = self._account_usage_query_history(start_date, end_date, self._dbs)
            else:
                query_text, params = self._info_schema_query_history(start_date, end_date, self._dbs)

            with ThreadSpinner(title='Pulling query history from Snowflake'):
                cursor.execute(query_text, params)
                logger.debug(f"Fetching results from Snowflake")
                rows = cursor.fetchall()

            rows_with_progress_bar = alive_it(rows, title="Parsing queries")
            for row in rows_with_progress_bar:
                query_context = QueryContext(queried_database=row[1],
                                             queried_schema=row[2],
                                             query_time=row[3],
                                             query_volume=row[4],
                                             query_type=row[5],
                                             user_name=row[6],
                                             role_name=row[7],
                                             duration=row[8],
                                             query_id=row[9],
                                             destination_table=row[10],
                                             referenced_tables=row[11])

                query = SnowflakeQuery(raw_query_text=row[0],
                                       query_context=query_context)

                self.add_query(query)
                # This is mainly used for debugging
                if query_context.destination_table is not None and query_context.referenced_tables is not None:
                    self.access_history_queries += 1
            logger.debug("Finished fetching snowflake history query results")

    def properties(self) -> dict:
        query_history_properties = super().properties()
        query_history_properties['query_history_properties'].update(
            {'query_history_source': self.query_history_source, 'access_history_queries': self.access_history_queries})
        return query_history_properties




import subprocess
from typing import Union

from utils.log import get_logger

logger = get_logger(__name__)


class DbtRunner(object):
    def __init__(self, project_dir: str, profiles_dir: str) -> None:
        self.project_dir = project_dir
        self.profiles_dir = profiles_dir

    def _run_command(self, command_args: list) -> bool:
        dbt_command = ['dbt']
        dbt_command.extend(command_args)
        dbt_command.extend(['--project-dir', self.project_dir])
        dbt_command.extend(['--profiles-dir', self.profiles_dir])
        logger.info(f"Running {' '.join(dbt_command)} (this might take a while)")
        result = subprocess.run(dbt_command, check=False, capture_output=True)
        logger.info(result.stdout.decode('utf-8'))
        if result.returncode != 0:
            return False

        return True

    def deps(self) -> bool:
        return self._run_command(['deps'])

    def seed(self) -> bool:
        return self._run_command(['seed'])

    def snapshot(self) -> bool:
        return self._run_command(['snapshot'])

    def run(self, model: Union[str, None] = None, full_refresh: bool = False) -> bool:
        command_args = ['run']
        if full_refresh:
            command_args.append('--full-refresh')
        if model is not None:
            command_args.extend(['-m', model])
        return self._run_command(command_args)




from datetime import datetime
from dateutil import tz

MILLISECONDS_IN_SEC = 1000
MILLISECONDS_IN_MIN = (1000 * 60)
MILLISECONDS_IN_HOUR = (1000 * 60 * 60)


def convert_utc_time_to_local_time(utc_time: 'datetime') -> 'datetime':
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    utc_time = utc_time.replace(tzinfo=from_zone)
    return utc_time.astimezone(to_zone)


def format_milliseconds(duration: int) -> str:

    seconds = int((duration / MILLISECONDS_IN_SEC) % 60)
    minutes = int((duration / MILLISECONDS_IN_MIN) % 60)
    hours = int(duration / MILLISECONDS_IN_HOUR)

    remaining_milliseconds = duration - (hours * MILLISECONDS_IN_HOUR + minutes * MILLISECONDS_IN_MIN +
                                         seconds * MILLISECONDS_IN_SEC)

    return f'{hours}h:{minutes}m:{seconds}s:{remaining_milliseconds}ms'


from alive_progress import alive_it
from google.cloud import bigquery

from lineage.bigquery_query import BigQueryQuery
from lineage.query import Query
from lineage.query_context import QueryContext
from lineage.query_history import QueryHistory
from utils.log import get_logger
from datetime import datetime

from utils.thread_spinner import ThreadSpinner

logger = get_logger(__name__)


class BigQueryQueryHistory(QueryHistory):
    PLATFORM_TYPE = 'bigquery'

    INFORMATION_SCHEMA_QUERY_HISTORY = """
    {database_name_normalized}_query_history as (
    SELECT query, end_time, dml_statistics.inserted_row_count + dml_statistics.updated_row_count, statement_type, 
    user_email, destination_table, referenced_tables, TIMESTAMP_DIFF(end_time, start_time, MILLISECOND),
    job_id
           
    FROM {database_name}.region-{location}.INFORMATION_SCHEMA.JOBS_BY_PROJECT
    WHERE
         project_id = '{database_name}'
         AND creation_time BETWEEN @start_time AND {creation_time_range_end_expr}
         AND end_time BETWEEN @start_time AND {end_time_range_end_expr}
         AND job_type = "QUERY"
         AND state = "DONE"
         AND error_result is NULL
         AND query NOT like '%JOBS_BY_PROJECT%'
    ),
    """
    INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP = 'CURRENT_TIMESTAMP()'
    INFO_SCHEMA_END_TIME_UP_TO_PARAMETER = '@end_time'

    SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY = "select * from {database_name_normalized}_query_history"

    UNION_ALL_DBS = """
        union_all_dbs as (
            {union_all_dbs}
        )
        select * from union_all_dbs
        order by end_time
    """

    def __init__(self, con, dbs: str, should_export_query_history: bool = True, full_table_names: bool = True) -> None:
        super().__init__(con, dbs, should_export_query_history, full_table_names)

    @classmethod
    def _build_history_query(cls, start_date: datetime, end_date: datetime, dbs: list, location: str) -> \
            (str, []):
        params = [bigquery.ScalarQueryParameter("start_time", "TIMESTAMP", start_date)]

        end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP
        creation_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP
        if end_date is not None:
            params.append(bigquery.ScalarQueryParameter("end_time",
                                                        "TIMESTAMP",
                                                        cls._include_end_date(end_date)))
            end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_PARAMETER
            creation_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_PARAMETER

        query_text = 'with'
        for db in dbs:
            query_text += cls.INFORMATION_SCHEMA_QUERY_HISTORY.\
                format(database_name_normalized=cls._normalize_database_name(db),
                       database_name=db,
                       location=location,
                       creation_time_range_end_expr=creation_time_range_end_expr,
                       end_time_range_end_expr=end_time_range_end_expr)

        dbs_count = len(dbs)
        union_all_dbs = cls.SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY.\
            format(database_name_normalized=cls._normalize_database_name(dbs[0]))
        for i in range(1, dbs_count):
            union_all_dbs += ' union all ' + cls.SELECT_FROM_INFORMATION_SCHEMA_QUERY_HISTORY. \
                format(database_name_normalized=cls._normalize_database_name(dbs[i]))

        query_text += cls.UNION_ALL_DBS.format(union_all_dbs=union_all_dbs)

        return query_text, params

    def _query_history_table(self, start_date: datetime, end_date: datetime) -> [Query]:
        logger.debug(f"Pulling BigQuery history from databases - {self._dbs}")

        query_text, query_parameters = self._build_history_query(start_date, end_date, self._dbs, self._con.location)

        job_config = bigquery.QueryJobConfig(
            query_parameters=query_parameters
        )

        with ThreadSpinner(title='Pulling query history from BigQuery'):
            job = self._con.query(query_text, job_config=job_config)
            logger.debug("Finished executing bigquery jobs history query")
            rows = list(job.result())

        rows_with_progress_bar = alive_it(rows, title="Parsing queries")
        for row in rows_with_progress_bar:
            query_context = QueryContext(query_time=row[1],
                                         query_volume=row[2],
                                         query_type=row[3],
                                         user_name=row[4],
                                         destination_table=row[5],
                                         referenced_tables=row[6],
                                         duration=row[7],
                                         query_id=row[8])

            query = BigQueryQuery(raw_query_text=row[0],
                                  query_context=query_context)

            self.add_query(query)

        logger.debug("Finished fetching bigquery history job results")




from ruamel.yaml import YAML


class OrderedYaml(object):

    def __init__(self) -> None:
        self.ordered_yaml = YAML()
        self.ordered_yaml.indent(mapping=2, sequence=4, offset=2)
        self.ordered_yaml.preserve_quotes = True

    def load(self, file_path: str) -> dict:
        with open(file_path, 'r') as file_obj:
            return self.ordered_yaml.load(file_obj)

    def dump(self, data: dict, file_path: str) -> dict:
        with open(file_path, 'w') as file_obj:
            return self.ordered_yaml.dump(data, file_obj)




import os


def is_flight_mode_on() -> bool:
    return is_env_var_on('FLIGHTMODE')


def is_debug_mode_on() -> bool:
    return is_env_var_on('DEBUG')


def is_env_var_on(env_var) -> bool:
    if os.getenv(env_var) == '1':
        print(env_var, ' is on!')
        return True

    return False



from setuptools import setup, find_packages
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='elementary-data',
    description='Data monitoring and lineage',
    version='0.1.4',
    packages=find_packages(),
    python_requires='>=3.6.2',
    entry_points='''
        [console_scripts]
        edr=cli.cli:cli
    ''',
    author="Elementary",
    package_data={"": ["header.html", "dbt_project/*"]},
    keyword="data, lineage, data lineage, data warehouse, DWH, observability, data monitoring, data observability, "
            "Snowflake, BigQuery, Redshift, data reliability, analytics engineering",
    long_description=README,
    install_requires=[
        'click>=8,<9',
        'pyfiglet',
        'snowflake-connector-python[secure-local-storage]>=2.4.1,<2.8.0',
        'dbt-core>=0.20,<1.1.0',
        'dbt-postgres>=0.20,<1.1.0',
        'dbt-redshift>=0.20,<1.1.0',
        'dbt-snowflake>=0.20,<1.1.0',
        'dbt-bigquery>=0.20,<1.1.0',
        'networkx>=2.3,<3',
        'sqllineage==1.2.4',
        'pyvis>=0.1,<1',
        'sqlparse>=0.3.1,<0.5',
        'alive-progress',
        'python-dateutil>=2.7.2,<3.0.0',
        'google-cloud-bigquery>=1.25.0,<3',
        'google-api-core>=1.16.0,<3',
        'requests>=2.7,<3.0.0',
        'beautifulsoup4',
        'posthog',
        'sqlfluff',
        'ruamel.yaml',
        # Indirect
        'protobuf<4,>=3.13.0',
        'typing-extensions<3.11,>=3.7.4',
        'flask<1.1.1',
        'Jinja2>=2.10.1,<3',
        'tomli<2.0.0,>=0.2.6',
        'urllib3<1.26,>=1.20'
    ],
    long_description_content_type="text/markdown",
    license='',
    url='https://github.com/elementary-data/elementary',
    author_email='or@elementary-data.com',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)


from config.config import Config
from lineage.bigquery_query_history import BigQueryQueryHistory
from utils.dbt import get_bigquery_client, get_snowflake_client
from exceptions.exceptions import ConfigError
from lineage.query_history import QueryHistory
from lineage.snowflake_query_history import SnowflakeQueryHistory


class QueryHistoryFactory(object):
    @staticmethod
    def create_query_history(config: 'Config', dbs: str) -> QueryHistory:
        if config.platform == 'snowflake':
            snowflake_con = get_snowflake_client(config.credentials, server_side_binding=False)
            return SnowflakeQueryHistory(con=snowflake_con, dbs=dbs, query_history_source=config.query_history_source)
        elif config.platform == 'bigquery':
            bigquery_client = get_bigquery_client(config.credentials)
            return BigQueryQueryHistory(con=bigquery_client, dbs=dbs)
        else:
            raise ConfigError("Unsupported profile type")



import os
import uuid
from typing import Union

import posthog
import requests
from bs4 import BeautifulSoup

from config.config import Config


class AnonymousTracking(object):
    ANONYMOUS_USER_ID_FILE = '.user_id'
    FETCH_API_KEY_AND_URL = 'https://www.elementary-data.com/telemetry'

    def __init__(self, config: Config) -> None:
        self.anonymous_user_id = None
        self.api_key = None
        self.url = None
        self.config = config
        self.do_not_track = config.anonymous_tracking_enabled is False
        self.run_id = str(uuid.uuid4())
        self.init()

    def init(self):
        legacy_anonymous_user_id_file_name = os.path.join(self.config.profiles_dir, self.ANONYMOUS_USER_ID_FILE)
        anonymous_user_id_file_name = os.path.join(self.config.config_dir, self.ANONYMOUS_USER_ID_FILE)
        # First check legacy file path
        if os.path.exists(legacy_anonymous_user_id_file_name):
            with open(legacy_anonymous_user_id_file_name, 'r') as anonymous_user_id_file:
                self.anonymous_user_id = anonymous_user_id_file.read()
        elif os.path.exists(anonymous_user_id_file_name):
            with open(anonymous_user_id_file_name, 'r') as anonymous_user_id_file:
                self.anonymous_user_id = anonymous_user_id_file.read()
        else:
            self.anonymous_user_id = str(uuid.uuid4())
            with open(anonymous_user_id_file_name, 'w') as anonymous_user_id_file:
                anonymous_user_id_file.write(self.anonymous_user_id)

        self.api_key, self.url = self._fetch_api_key_and_url()

    @classmethod
    def _fetch_api_key_and_url(cls) -> (Union[str, None], Union[str, None]):
        result = requests.get(url=cls.FETCH_API_KEY_AND_URL)
        if result.status_code != 200:
            return None, None
        soup = BeautifulSoup(result.content, 'html.parser')
        h5_tag = soup.find('h5')
        if h5_tag is None:
            return None, None

        api_key_and_url = h5_tag.text.split('\n')
        if len(api_key_and_url) != 2:
            return None, None

        return api_key_and_url[0], api_key_and_url[1]

    def send_event(self, name: str, properties: dict = None) -> None:
        if self.do_not_track:
            return

        if self.api_key is None or self.url is None or self.anonymous_user_id is None:
            return

        if properties is None:
            properties = dict()

        properties['run_id'] = self.run_id
        properties['platform'] = self.config.platform

        posthog.api_key = self.api_key
        posthog.host = self.url
        posthog.capture(distinct_id=self.anonymous_user_id, event=name, properties=properties)


def track_cli_start(anonymous_tracking: AnonymousTracking, module_name: str, cli_properties: dict,
                    command: str = None) -> None:
    try:
        cli_start_properties = {'cli_properties': cli_properties,
                                'module_name': module_name,
                                'command': command}
        anonymous_tracking.send_event('cli-start', properties=cli_start_properties)
    except Exception:
        pass


def track_cli_end(anonymous_tracking: AnonymousTracking, module_name: str, execution_properties: dict,
                  command: str = None) -> None:
    try:
        if anonymous_tracking is None:
            return

        cli_end_properties = {'execution_properties': execution_properties,
                              'module_name': module_name,
                              'command': command}
        anonymous_tracking.send_event('cli-end', properties=cli_end_properties)
    except Exception:
        pass


def track_cli_exception(anonymous_tracking: AnonymousTracking, module_name: str, exc: Exception,
                        command: str = None) -> None:
    try:
        if anonymous_tracking is None:
            return

        cli_exception_properties = {'exception_properties': {'exception_type': str(type(exc)),
                                                             'exception_content': str(exc)},
                                    'module_name': module_name,
                                    'command': command}
        anonymous_tracking.send_event('cli-exception', properties=cli_exception_properties)
    except Exception:
        pass


def track_cli_help(anonymous_tracking: AnonymousTracking):
    try:
        anonymous_tracking.send_event('cli-help')
    except Exception:
        pass